#pragma once

#include <dune/common/typeutilities.hh>

#include <amdis/LinearAlgebra.hpp>
#include <amdis/OperatorList.hpp>
#include <amdis/common/FlatVector.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/typetree/TreePath.hpp>

namespace AMDiS
{
  /// \brief The basic container that stores a base vector and a corresponding basis
  /**
   * A vector storing all the assembled Operators indexed with DOF indices. The vector
   * data is associated to a global basis.
   *
   * \tparam GB  Basis of the vector
   * \tparam T   Coefficient type to store in the vector
   **/
  template <class GB, class T = double>
  class LinearForm
      : public VectorBase<GB, VectorBackend<BackendTraits<GB,T>>>
  {
    using Super = VectorBase<GB, VectorBackend<BackendTraits<GB,T>>>;
    using Self = LinearForm;

  public:
    using Backend = VectorBackend<BackendTraits<GB,T>>;

    /// The type of the functionspace basis associated to this linearform
    using GlobalBasis = GB;
    using LocalView = typename GlobalBasis::LocalView;

    /// A traits class collecting several parameters of the linear algebra backend
    using Traits = typename Backend::Traits;

    /// The type of the elements of the DOFVector
    using CoefficientType = typename Traits::CoefficientType;

    /// The type of the vector filled on an element with local contributions
    using ElementVector = FlatVector<CoefficientType>;

  public:
    /// Constructor. Stores the shared_ptr of the basis and creates a new DataTransfer.
    template <class GB_,
      Dune::disableCopyMove<Self, GB_> = 0>
    explicit LinearForm(GB_&& basis)
      : Super(FWD(basis))
    {
      operators_.init(*Super::basis());
    }

    /// Prepare the LinearForm for insertion of values, finish the insertion with
    /// \ref finish().
    void init(bool clear)
    {
      Super::init(clear);

      auto const localSize = this->basis()->localView().maxSize();
      elementVector_.resize(localSize);
    }

    /// \brief Associate a local operator with this LinearForm
    /**
     * Stores an operator in a list that gets assembled during a call to \ref assemble().
     * The operator may be assigned to a specific context, i.e. either an element
     * operator, an intersection operator, or a boundary operator.
     * The \p path tree paths specify the sub-basis for test functions the operator
     * is applied to.
     *
     * \tparam ContextTag  One of \ref tag::element_operator, \ref tag::intersection_operator
     *                     or \ref tag::boundary_operator indicating where to assemble
     *                     this operator.
     * \tparam Expr        An pre-operator that can be bound to a gridView, or a valid
     *                     GridOperator.
     * \tparam path        A tree-path for the Basis
     *
     * [[expects: path is valid tree-path in Basis]]
     * @{
     **/
    template <class ContextTag, class Expr, class TreePath>
    void addOperator(ContextTag contextTag, Expr const& expr, TreePath path);

    // Add an operator to be assembled on the elements of the grid
    template <class Expr, class TreePath = RootTreePath>
    void addOperator(Expr const& expr, TreePath path = {})
    {
      using E = typename LocalView::Element;
      addOperator(tag::element_operator<E>{}, expr, path);
    }

    // Add an operator to be assembled on the intersections of the grid
    template <class Expr, class TreePath = RootTreePath>
    void addIntersectionOperator(Expr const& expr, TreePath path = {})
    {
      using I = typename LocalView::GridView::Intersection;
      addOperator(tag::intersection_operator<I>{}, expr, path);
    }
    /// @}


    /// Assemble the vector operators on the bound element.
    void assemble(LocalView const& localView);

    /// Assemble all vector operators added by \ref addOperator().
    void assemble();

  private:
    /// Dense vector to store coefficients during \ref assemble()
    ElementVector elementVector_;

    /// List of operators associated to nodes, filled in \ref addOperator().
    VectorOperators<GlobalBasis,ElementVector> operators_;
  };


#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  template <class GB>
  LinearForm(GB&& basis)
    -> LinearForm<Underlying_t<GB>>;
#endif

  template <class T = double, class GB>
  auto makeLinearForm(GB&& basis)
  {
    using LF = LinearForm<Underlying_t<GB>, T>;
    return LF{FWD(basis)};
  }

} // end namespace AMDiS

#include <amdis/LinearForm.inc.hpp>
