#pragma once

#include <limits>

#include <dune/common/reservedvector.hh>
#include <amdis/LinearAlgebra.hpp>
#include <amdis/functions/FunctionFromCallable.hpp>

namespace AMDiS {

template <class D, class MI>
  template <class RB, class CB>
void PeriodicBC<D,MI>::
init(RB const& basis, CB const& /*colBasis*/)
{
  if (!bool(hasConnectedIntersections_)) {
    hasConnectedIntersections_ = true;
    const auto& gridView = basis.gridView().grid().levelGridView(0);
    for (auto const& element : elements(gridView)) {
      for (const auto& intersection : intersections(gridView, element)) {
        if (!this->onBoundary(intersection))
          continue;

        if (!intersection.neighbor()) {
          hasConnectedIntersections_ = false;
          break;
        }
      }
    }
  }

  if (*hasConnectedIntersections_)
    initAssociations(basis);
  else
    initAssociations2(basis);
}


template <class D, class MI>
  template <class Basis>
void PeriodicBC<D,MI>::
initAssociations(Basis const& basis)
{
  using std::sqrt;
  static const auto tol = sqrt(std::numeric_limits<typename D::field_type>::epsilon());
  periodicNodes_.clear();
  periodicNodes_.resize(basis.dimension(), false);

  auto localView = basis.localView();
  auto seDOFs = subEntityDOFs(basis);
  const auto& gridView = basis.gridView();
  for (auto const& element : elements(gridView)) {
    if (!element.hasBoundaryIntersections())
      continue;

    localView.bind(element);
    for (const auto& intersection : intersections(gridView, element)) {
      if (!this->onBoundary(intersection))
        continue;

      test_exit(intersection.neighbor(),
        "Neighbors of periodic intersection not assigned. Use a periodic Grid instead!");

      // collect DOFs of inside element on intersection
      localView.bind(intersection.inside());
      seDOFs.bind(localView, intersection.indexInInside(), 1);
      std::vector<std::size_t> insideDOFs(seDOFs.begin(), seDOFs.end());
      std::vector<MI> insideGlobalDOFs(insideDOFs.size());
      std::transform(insideDOFs.begin(), insideDOFs.end(), insideGlobalDOFs.begin(),
        [&](std::size_t localIndex) { return localView.index(localIndex); });
      auto insideCoords = coords(localView.tree(), insideDOFs);

      // collect DOFs of ouside element on intersection
      localView.bind(intersection.outside());
      seDOFs.bind(localView, intersection.indexInOutside(), 1);
      std::vector<std::size_t> outsideDOFs(seDOFs.begin(), seDOFs.end());
      auto outsideCoords = coords(localView.tree(), outsideDOFs);

      // compare mapped coords of DOFs
      assert(insideDOFs.size() == outsideDOFs.size());
      for (std::size_t i = 0; i < insideCoords.size(); ++i) {
        auto x = faceTrafo_.evaluate(insideCoords[i]);
        for (std::size_t j = 0; j < outsideCoords.size(); ++j) {
          auto const& y = outsideCoords[j];
          if (distance(x,y) < tol) {
            periodicNodes_[insideGlobalDOFs[i]] = true;
            associations_[insideGlobalDOFs[i]] = localView.index(outsideDOFs[j]);
          }
        }
      }
    }
  }
}

namespace Impl
{
  template <class D>
  class CompareTol
  {
    using ctype = typename D::field_type;

  public:
    CompareTol(ctype tol)
      : tol_(tol)
    {}

    bool operator()(D const& lhs, D const& rhs) const
    {
      using std::abs;
      for (int i = 0; i < D::dimension; i++) {
        if (abs(lhs[i] - rhs[i]) < tol_)
          continue;
        return lhs[i] < rhs[i];
      }
      return false;
    }

  private:
    ctype tol_;
  };
}



template <class D, class MI>
  template <class Basis>
void PeriodicBC<D,MI>::
initAssociations2(Basis const& basis)
{
  using std::sqrt;
  static const auto tol = sqrt(std::numeric_limits<typename D::field_type>::epsilon());
  periodicNodes_.clear();
  periodicNodes_.resize(basis.dimension(), false);

  // Dune::ReservedVector<std::pair<EntitySeed,int>,2>
  using EntitySeed = typename Basis::GridView::Grid::template Codim<0>::EntitySeed;
  using CoordAssoc = std::map<D, std::vector<std::pair<EntitySeed,int>>, Impl::CompareTol<D>>;
  CoordAssoc coordAssoc(Impl::CompareTol<D>{tol});

  // generate periodic element pairs
  const auto& gridView = basis.gridView();
  for (auto const& element : elements(gridView)) {
    for (const auto& intersection : intersections(gridView, element)) {
      if (!this->onBoundary(intersection))
        continue;

      auto x = intersection.geometry().center();
      auto seed = intersection.inside().seed();
      coordAssoc[x].push_back({seed,intersection.indexInInside()});
      coordAssoc[faceTrafo_.evaluate(x)].push_back({seed,intersection.indexInInside()});
    }
  }

  auto localView = basis.localView();
  auto seDOFs = subEntityDOFs(basis);
  for (auto const& assoc : coordAssoc) {
    auto const& seeds = assoc.second;
    if (seeds.size() != 2)
      continue;

    // collect DOFs of inside element on intersection
    auto el1 = gridView.grid().entity(seeds[0].first);
    localView.bind(el1);
    seDOFs.bind(localView, seeds[0].second, 1);
    std::vector<std::size_t> insideDOFs(seDOFs.begin(), seDOFs.end());
    std::vector<MI> insideGlobalDOFs(insideDOFs.size());
    std::transform(insideDOFs.begin(), insideDOFs.end(), insideGlobalDOFs.begin(),
      [&](std::size_t localIndex) { return localView.index(localIndex); });
    auto insideCoords = coords(localView.tree(), insideDOFs);

    // collect DOFs of ouside element on intersection
    auto el2 = gridView.grid().entity(seeds[1].first);
    localView.bind(el2);
    seDOFs.bind(localView, seeds[1].second, 1);
    std::vector<std::size_t> outsideDOFs(seDOFs.begin(), seDOFs.end());
    auto outsideCoords = coords(localView.tree(), outsideDOFs);

    // compare mapped coords of DOFs
    assert(insideDOFs.size() == outsideDOFs.size());
    for (std::size_t i = 0; i < insideCoords.size(); ++i) {
      auto x = faceTrafo_.evaluate(insideCoords[i]);
      for (std::size_t j = 0; j < outsideCoords.size(); ++j) {
        auto const& y = outsideCoords[j];
        if (distance(x,y) < tol) {
          periodicNodes_[insideGlobalDOFs[i]] = true;
          associations_[insideGlobalDOFs[i]] = localView.index(outsideDOFs[j]);
        }
      }
    }
  }
}


template <class D, class MI>
  template <class Node>
std::vector<D> PeriodicBC<D,MI>::
coords(Node const& tree, std::vector<std::size_t> const& localIndices) const
{
  std::vector<D> dofCoords(localIndices.size());
  for_each_leaf_node(tree, [&](auto const& node, auto const& tp)
  {
    std::size_t size = node.finiteElement().size();
    auto geometry = node.element().geometry();

    auto const& localInterpol = node.finiteElement().localInterpolation();
    using FiniteElement = TYPEOF(node.finiteElement());
    using DomainType = typename FiniteElement::Traits::LocalBasisType::Traits::DomainType;
    using RangeType = typename FiniteElement::Traits::LocalBasisType::Traits::RangeType;

    std::array<std::vector<typename RangeType::field_type>, D::dimension> coeffs;
    for (int d = 0; d < D::dimension; ++d) {
      auto evalCoord = functionFromCallable<RangeType(DomainType)>([&](DomainType const& local) -> RangeType { return geometry.global(local)[d]; });
      localInterpol.interpolate(evalCoord, coeffs[d]);
    }

    for (std::size_t j = 0; j < localIndices.size(); ++j) {
      D x;
      for (std::size_t i = 0; i < size; ++i) {
        if (node.localIndex(i) == localIndices[j]) {
          for (int d = 0; d < D::dimension; ++d)
            x[d] = coeffs[d][i];
          break;
        }
      }
      dofCoords[j] = std::move(x);
    }
  });

  return dofCoords;
}


template <class D, class MI>
  template <class Mat, class Sol, class Rhs, class RN, class RTP, class CN, class CTP>
void PeriodicBC<D,MI>::
fillBoundaryCondition(Mat& matrix, Sol& solution, Rhs& rhs, RN const& rowNode, RTP rowTreePath, CN const& colNode, CTP colTreePath)
{
  periodicBC(matrix, solution, rhs, periodicNodes_, associations_);
}

} // end namespace AMDiS
