#pragma once

#include <type_traits>

#include <dune/grid/utility/hierarchicsearch.hh>

#include <amdis/functions/Interpolate.hpp>
#include <amdis/gridfunctions/GridFunction.hpp>

namespace AMDiS {

// Evaluate DiscreteFunction in global coordinates
template <class GB, class VT, class TP>
typename DiscreteFunction<GB,VT,TP,true>::Range DiscreteFunction<GB,VT,TP,true>::
  operator()(Domain const& x) const
{
  using Grid = typename GlobalBasis::GridView::Grid;
  using IS = typename GlobalBasis::GridView::IndexSet;

  auto const& gv = this->basis()->gridView();
  Dune::HierarchicSearch<Grid,IS> hsearch{gv.grid(), gv.indexSet()};

  auto element = hsearch.findEntity(x);
  auto geometry = element.geometry();
  auto localFct = localFunction(*this);
  localFct.bind(element);
  return localFct(geometry.local(x));
}


// Interpolation of GridFunction to DOFVector
template <class GB, class VT, class TP>
  template <class Expr, class Tag>
void DiscreteFunction<GB,VT,TP,false>::
  interpolate_noalias(Expr&& expr, Tag strategy)
{
  auto const& basis = *this->basis();
  auto const& treePath = this->treePath();

  auto&& gf = makeGridFunction(FWD(expr), basis.gridView());

  if (std::is_same<Tag, tag::average>::value) {
    auto counter = coefficients();
    AMDiS::interpolate(basis, coefficients(), gf, treePath, counter);

    coefficients().forEach([&counter](std::size_t dof, auto& coeff)
    {
      coeff /= std::max(double(counter.at(dof)), 1.0);
    });
  } else {
    AMDiS::interpolate(basis, coefficients(), gf, treePath);
  }
}


// Interpolation of GridFunction to DOFVector
template <class GB, class VT, class TP>
  template <class Expr, class Tag>
void DiscreteFunction<GB,VT,TP,false>::
  interpolate(Expr&& expr, Tag strategy)
{
  // create temporary copy of data
  DOFVector<GB,VT> tmp(coefficients());

  Self tmpView{tmp, this->treePath()};
  tmpView.interpolate_noalias(FWD(expr), strategy);

  // move data from temporary vector into stored DOFVector
  coefficients().backend() = std::move(tmp.backend());
}

} // end namespace AMDiS
