#pragma once

#include <string>
#include <memory>

#if HAVE_DUNE_VTK

#include <dune/vtk/pvdwriter.hh>
#include <dune/vtk/vtkwriter.hh>

#include <amdis/Initfile.hpp>
#include <amdis/common/Filesystem.hpp>
#include <amdis/io/FileWriterBase.hpp>

namespace AMDiS
{
  /// Adapter for the dune-vtk writer
  template <class GB, class VT, class TP>
  class DuneVtkWriter
      : public FileWriterBase
  {
    using GridView = typename GB::GridView;
    using Writer = Dune::VtkWriter<GridView>;
    using SeqWriter = Dune::PvdWriter<Writer>;

  public:
    /// Constructor.
    DuneVtkWriter(std::string const& name, DiscreteFunction<GB,VT,TP> const& discreteFct)
      : FileWriterBase(name)
      , discreteFct_(discreteFct)
    {
      int m = 0, p = 0;
      Parameters::get(name + "->animation", animation_);
      Parameters::get(name + "->mode", m);
      Parameters::get(name + "->precision", p);

      Dune::Vtk::FormatTypes mode =
        m == 0 ? Dune::Vtk::ASCII :
        m == 1 ? Dune::Vtk::BINARY :
                 Dune::Vtk::COMPRESSED;

      Dune::Vtk::DataTypes precision =
        p == 0 ? Dune::Vtk::FLOAT32 :
                 Dune::Vtk::FLOAT64;

      if (animation_) {
        vtkSeqWriter_ = std::make_shared<SeqWriter>(gridView(), mode, precision);
        vtkSeqWriter_->addPointData(discreteFct_, this->name());
      } else {
        vtkWriter_ = std::make_shared<Writer>(gridView(), mode, precision);
        vtkWriter_->addPointData(discreteFct_, this->name());
      }
    }

    /// Implements \ref FileWriterBase::write
    void write(AdaptInfo& adaptInfo, bool force) override
    {
      std::string data_dir = this->dir() + "/_piecefiles";
      filesystem::create_directories(data_dir);

      std::string filename = filesystem::path({this->dir(), this->filename() + ".vtk"}).string();
      if (force || this->doWrite(adaptInfo)) {
        if (animation_)
          vtkSeqWriter_->writeTimestep(adaptInfo.time(), filename, data_dir, true);
        else
          vtkWriter_->write(filename, data_dir);
      }
    }

  private:
    /// The Gridview this writer lives on. Given by the discrete function.
    GridView const& gridView() const
    {
      return discreteFct_.basis()->gridView();
    }

  private:
    DiscreteFunction<GB,VT,TP> discreteFct_;

    std::shared_ptr<Writer> vtkWriter_;
    std::shared_ptr<SeqWriter> vtkSeqWriter_;

    /// write .pvd if animation=true, otherwise write only .vtu
    bool animation_ = false;
  };

} // end namespace AMDiS

#endif // HAVE_DUNE_VTK
