#pragma once

#include <string>

#include <dune/common/timer.hh>

// AMDiS includes
#include <amdis/CreatorInterface.hpp>
#include <amdis/Output.hpp>
#include <amdis/linearalgebra/LinearSolverInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>

namespace AMDiS
{
  /** \ingroup Solver
   *
   * \brief
   * Wrapper class for various solvers. These solvers
   * are parametrized by MatrixType and VectorType. The different
   * algorithms, like krylov subspace methods or other external
   * solvers where the backend provides an interface, can be assigned
   * by different Runner objects.
   **/
  template <class Traits, class Runner>
  class LinearSolver
      : public LinearSolverInterface<Traits>
  {
    using Self = LinearSolver;
    using Super = LinearSolverInterface<Traits>;

    using Mat = typename Traits::Mat;
    using Vec = typename Traits::Vec;
    using Comm = typename Traits::Comm;

  public:
    /// A creator to be used instead of the constructor.
    struct Creator : CreatorInterfaceName<Super>
    {
      std::unique_ptr<Super> createWithString(std::string prefix) override
      {
        return std::make_unique<Self>(prefix);
      }
    };

  public:
    /// Constructor
    explicit LinearSolver(std::string prefix)
      : runner_(prefix)
    {}

  private:
    /// Implements \ref LinearSolverInterface::solveSystemImpl()
    void solveImpl(Mat const& A, Vec& x, Vec const& b, Comm& comm, SolverInfo& solverInfo) override
    {
      Dune::Timer t;
      if (solverInfo.doCreateMatrixData()) {
        // init matrix or wrap block-matrix or ...
        runner_.init(A, comm);
      }

      if (solverInfo.info() > 0)
        msg("prepare solver needed {} seconds", t.elapsed());

      int error = runner_.solve(A, x, b, solverInfo);
      solverInfo.setError(error);

      if (!solverInfo.doStoreMatrixData())
        runner_.exit();
    }

  private:
    /// redirect the implementation to a runner. Implementing a \ref RunnerInterface
    Runner runner_;
  };

} // end namespace AMDiS
