#pragma once

#include <memory>
#include <string>

#include <dune/common/std/optional.hh>

#include <dune/istl/owneroverlapcopy.hh>
#include <dune/istl/solvercategory.hh>
#include <dune/istl/paamg/pinfo.hh>

#include <amdis/Environment.hpp>
#include <amdis/functions/GlobalIdSet.hpp>
#include <amdis/linearalgebra/Communication.hpp>
#include <amdis/linearalgebra/ParallelIndexSet.hpp>

namespace AMDiS
{
  /// Dummy implementation for ISTL-specific communication when no MPI is found
  template <class Basis>
  class SequentialISTLCommunication
  {
    using Impl = Dune::Amg::SequentialInformation;

  public:
    using Sequential = SequentialISTLCommunication;

    SequentialISTLCommunication() = default;

    SequentialISTLCommunication(Basis const&, typename Dune::SolverCategory::Category cat)
    {
      assert(cat == Dune::SolverCategory::sequential);
    }

    typename Dune::SolverCategory::Category category() const
    {
      return Dune::SolverCategory::sequential;
    }

    Impl const& get() const
    {
      return impl_;
    }

    Sequential const& sequential() const
    {
      return *this;
    }

    void update(Basis const&) { /* do nothing */ }

  private:
    Impl impl_;
  };


#if HAVE_MPI
  /// Implementation class for ISTL-specific communication to be used in parallel solvers
  template <class Basis>
  class ISTLCommunication
  {
    using GlobalId = typename GlobalBasisIdSet<Basis>::IdType;
    using LocalIndex = typename Basis::size_type;
    using Impl = Dune::OwnerOverlapCopyCommunication<GlobalId, LocalIndex>;

  public:
    using Sequential = SequentialISTLCommunication<Basis>;
    using IndexSet = typename Impl::ParallelIndexSet;
    using RemoteIndices = typename Impl::RemoteIndices;

  public:
    ISTLCommunication(Basis const& basis, typename Dune::SolverCategory::Category cat)
      : cat_(cat)
    {
      update(basis);
    }

    IndexSet const& indexSet() const
    {
      assert(bool(impl_));
      return impl_->indexSet();
    }

    RemoteIndices const& remoteIndices() const
    {
      assert(bool(impl_));
      return impl_->remoteIndices();
    }

    typename Dune::SolverCategory::Category category() const
    {
      return cat_;
    }

    Impl const& get() const
    {
      assert(bool(impl_));
      return *impl_;
    }

    Sequential sequential() const
    {
      return Sequential{};
    }

    void update(Basis const& basis)
    {
      impl_.emplace(Environment::comm(), cat_);

      auto& pis = impl_->indexSet();
      buildParallelIndexSet(basis, pis);

      auto& ris = impl_->remoteIndices();
      ris.setIndexSets(pis, pis, impl_->communicator());
      ris.template rebuild<true>();
    }

  private:
    typename Dune::SolverCategory::Category cat_;
    Dune::Std::optional<Impl> impl_ = {};
  };

#else // !HAVE_MPI

  template <class Basis>
  using ISTLCommunication = SequentialISTLCommunication<Basis>;

#endif


  template <class Basis>
  struct CommunicationCreator<ISTLCommunication<Basis>>
  {
    using Communication = ISTLCommunication<Basis>;

    /**
     * ISTLCommunication creator function. Takes the global basis and the solver's initfile prefix
     * and returns a unique pointer to the ISTLCommunication object.
     * Deduces the category of the solver used from the initfile and basis
     *
     * \param basis   Global basis associated to the solvers' matrices and vectors
     * \param prefix  Solver prefix used in the initfile, usually "<problem name>->solver"
     **/
    static std::unique_ptr<Communication> create(Basis const& basis, std::string const& prefix = "");
  };

} // end namespace AMDiS

#include <amdis/linearalgebra/istl/Communication.inc.hpp>
