#pragma once

#include <petsc.h>

#include <dune/common/parallel/indexset.hh>
#include <dune/common/parallel/localindex.hh>
#include <dune/common/parallel/plocalindex.hh>
#include <dune/common/parallel/remoteindices.hh>
#include <dune/common/std/optional.hh>

#include <amdis/functions/GlobalIdSet.hpp>
#include <amdis/linearalgebra/Communication.hpp>
#include <amdis/linearalgebra/DOFMapping.hpp>
#include <amdis/linearalgebra/ParallelIndexSet.hpp>

namespace AMDiS
{
  template <class Basis>
  class DistributedCommunication
  {
  public:
    using GlobalIdType = typename GlobalBasisIdSet<Basis>::IdType;
#if HAVE_MPI
    using LocalIndexType = Dune::ParallelLocalIndex<DefaultAttributeSet::Type>;
    using IndexSet = Dune::ParallelIndexSet<GlobalIdType, LocalIndexType, 512>;
    using RemoteIndices = Dune::RemoteIndices<IndexSet>;
#else
    struct IndexSet
    {
      constexpr std::size_t size() const { return size_; }
      std::size_t size_ = 0;
    };

    struct RemoteIndices {};
#endif
    using DofMap = DofMapping<IndexSet, PetscInt>;

  public:
    DistributedCommunication(Basis const& basis)
    {
      update(basis);
    }

    /// Return the ParallelIndexSet
    IndexSet const& indexSet() const { return *indexSet_; }
    IndexSet&       indexSet()       { return *indexSet_; }

    /// Return the RemoteIndices
    RemoteIndices const& remoteIndices() const { return remoteIndices_; }
    RemoteIndices&       remoteIndices()       { return remoteIndices_; }

    /// Return the DofMapping
    DofMap const& dofMap() const { return dofMap_; }
    DofMap&       dofMap()       { return dofMap_; }

    /// Update the indexSet, remoteIndices and dofmapping
    void update(Basis const& basis)
    {
      indexSet_.emplace();

#if HAVE_MPI
      buildParallelIndexSet(basis, *indexSet_);
      remoteIndices_.setIndexSets(*indexSet_, *indexSet_, Environment::comm());
      remoteIndices_.template rebuild<true>();
#else
      (*indexSet_).size_ = basis.dimension();
#endif

      dofMap_.update(*this);
    }

  protected:
    Dune::Std::optional<IndexSet> indexSet_;
    RemoteIndices remoteIndices_;
    DofMap dofMap_;
  };


  template <class Basis>
  struct CommunicationCreator<DistributedCommunication<Basis>>
  {
    using Communication = DistributedCommunication<Basis>;

    static std::unique_ptr<Communication> create(Basis const& basis, std::string const& prefix = "")
    {
      return std::make_unique<Communication>(basis);
    }
  };

} // end namespace AMDiS
