#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_gradtrial {};
    struct grad_trial {};
  }


  /// first-order operator \f$ \langle\psi, \mathbf{b}\cdot\nabla\phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::test_gradtrial, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_gradtrial, LC, GridFct>, LC, GridFct>
  {
    static const int dow = ContextGeometry<LC>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( static_size_v<typename GridFct::Range> == dow, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::test_gradtrial, GridFct const& expr)
      : Super(expr, 1)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RangeFieldType = typename NodeQuadCache<CN>::Traits::RangeFieldType;

      NodeQuadCache<RN> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<CN> colCache(colLocalFE.localBasis());

      auto const& shapeValuesCache = rowCache.evaluateFunctionAtQP(contextGeo, quad);
      auto const& shapeGradientsCache = colCache.evaluateJacobianAtQP(contextGeo, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = contextGeo.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto b = Super::coefficient(local);

        // the values of the shape functions on the reference element at the quadrature point
        auto const& shapeValues = shapeValuesCache[iq];

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,CG::dow>;
        thread_local std::vector<WorldVector> colGradients;
        colGradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < colGradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], colGradients[i]);

        for (std::size_t j = 0; j < colSize; ++j) {
          const auto local_j = colNode.localIndex(j);
          const auto value = factor * (b * colGradients[j]);
          for (std::size_t i = 0; i < rowSize; ++i) {
            const auto local_i = rowNode.localIndex(i);
            elementMatrix[local_i][local_j] += value * shapeValues[i];
          }
        }
      }

    }
  };


  /// Create a first-order term with derivative on test-function
  template <class Expr, class... QuadratureArgs>
  auto fot(Expr&& expr, tag::grad_trial, QuadratureArgs&&... args)
  {
    return makeOperator(tag::test_gradtrial{}, FWD(expr), FWD(args)...);
  }

  /** @} **/

} // end namespace AMDiS
