#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>
#include <amdis/utility/LocalToGlobalAdapter.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct partialtest_partialtrial
    {
      std::size_t comp_test; // i
      std::size_t comp_trial; // j
    };
  }


  /// second-order operator \f$ \langle\partial_i\psi, c\,\partial_j\phi\rangle \f$
  template <class LC, class GridFct>
  class GridFunctionOperator<tag::partialtest_partialtrial, LC, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::partialtest_partialtrial, LC, GridFct>, LC, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LC, GridFct>;

    static_assert( static_size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::partialtest_partialtrial tag, GridFct const& expr)
      : Super(expr, 2)
      , compTest_(tag.comp_test)
      , compTrial_(tag.comp_trial)
    {}

    template <class CG, class RN, class CN, class Mat>
    void getElementMatrix(CG const& contextGeo, RN const& rowNode, CN const& colNode, Mat& elementMatrix)
    {
      static_assert(RN::isLeaf && CN::isLeaf, "Operator can be applied to Leaf-Nodes only.");

      auto const& quad = this->getQuadratureRule(contextGeo.type(), rowNode, colNode);

      auto rowLocalBasis = makeLocalToGlobalBasisAdapter(rowNode, contextGeo.geometry());
      auto colLocalBasis = makeLocalToGlobalBasisAdapter(colNode, contextGeo.geometry());

      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = contextGeo.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        auto factor = contextGeo.integrationElement(quad[iq].position()) * quad[iq].weight();
        factor *= Super::coefficient(local);

        auto const& rowPartial = rowLocalBasis.partialsAt(local, compTest_);
        auto const& colPartial = colLocalBasis.partialsAt(local, compTrial_);

        for (std::size_t j = 0; j < colLocalBasis.size(); ++j) {
          const auto local_j = colNode.localIndex(j);
          const auto value = factor * colPartial[j];
          for (std::size_t i = 0; i < rowLocalBasis.size(); ++i) {
            const auto local_i = rowNode.localIndex(i);
            elementMatrix[local_i][local_j] += value * rowPartial[i];
          }
        }
      }

    }

  private:
    std::size_t compTest_;
    std::size_t compTrial_;
  };


  /// Create a second-order term of partial derivatives
  template <class Expr, class... QuadratureArgs>
  auto sot_ij(Expr&& expr, std::size_t comp_test, std::size_t comp_trial, QuadratureArgs&&... args)
  {
    return makeOperator(tag::partialtest_partialtrial{comp_test, comp_trial}, FWD(expr), FWD(args)...);
  }

  /** @} **/

} // end namespace AMDiS
