#pragma once

#include <dune/common/hybridutilities.hh>
#include <dune/common/rangeutilities.hh>

#include <dune/typetree/childextraction.hh>
#include <dune/typetree/nodetags.hh>
#include <dune/typetree/treepath.hh>
#include <dune/typetree/visitor.hh>

#include <amdis/common/ForEach.hpp>
#include <amdis/common/Logical.hpp>
#include <amdis/common/Range.hpp>
#include <amdis/common/TypeTraits.hpp>

// NOTE: backport of dune/typetree/traversal.hpp from Dune 2.7

namespace AMDiS {

    enum class TreePathType
    {
      DYNAMIC, STATIC
    };

    namespace Impl {

      // This is a constexpr version of the ternery operator c?t1:t1.
      // In contrast to the latter the type of t1 and t2 can be different.
      // Notice that std::conditional would not do the trick, because
      // it only selects between types.
      template<bool c, class T1, class T2,
        std::enable_if_t<c, int> = 0>
      constexpr auto conditionalValue(T1&& t1, T2&& t2) {
        return std::forward<T1>(t1);
      }

      template<bool c, class T1, class T2,
        std::enable_if_t<not c, int> = 0>
      constexpr auto conditionalValue(T1&& t1, T2&& t2) {
        return std::forward<T2>(t2);
      }


      /* The signature is the same as for the public applyToTree
       * function in Dune::Typetree, despite the additionally passed
       * treePath argument. The path passed here is associated to
       * the tree and the relative paths of the children (wrt. to tree)
       * are appended to this.  Hence the behavior of the public function
       * is resembled by passing an empty treePath.
       */

      /*
       * This is the overload for leaf traversal
       */
      template<class T, class TP, class V,
        std::enable_if_t<remove_cvref_t<T>::isLeaf, int> = 0>
      void apply_to_tree(T&& tree, TP treePath, V&& visitor)
      {
        visitor.leaf(tree, treePath);
      }

      /*
       * This is the general overload doing child traversal.
       */
      template<class T, class TP, class V,
        std::enable_if_t<not remove_cvref_t<T>::isLeaf, int> = 0>
      void apply_to_tree(T&& tree, TP treePath, V&& visitor)
      {
        // Do we really want to take care for const-ness of the Tree
        // when instantiating VisitChild below? I'd rather expect this:
        // using Tree = remove_cvref_t<T>;
        // using Visitor = remove_cvref_t<V>;
        using Tree = std::remove_reference_t<T>;
        using Visitor = std::remove_reference_t<V>;
        visitor.pre(tree, treePath);

        // Use statically encoded degree unless tree
        // is a power node and dynamic traversal is requested.
        constexpr auto useDynamicTraversal = (Tree::isPower and Visitor::treePathType==TreePathType::DYNAMIC);
        auto degree = conditionalValue<useDynamicTraversal>(Tree::degree(), Dune::index_constant<Tree::degree()>{});

        auto indices = Dune::range(degree);
        Dune::Hybrid::forEach(indices, [&](auto i) {
          auto childTP = Dune::TypeTree::push_back(treePath, i);
          auto&& child = tree.child(i);
          using Child = TYPEOF(child);

          visitor.beforeChild(tree, child, treePath, i);

          // This requires that visiotor.in(...) can always be instantiated,
          // even if there's a single child only.
          if (i>0)
            visitor.in(tree, treePath);
          static constexpr auto visitChild = Visitor::template VisitChild<Tree,Child,TP>::value;
          #if AMDIS_HAS_CXX_CONSTEXPR_IF
          if constexpr(visitChild)
            applyToTree(child, childTP, visitor);
          #else // AMDIS_HAS_CXX_CONSTEXPR_IF
          Dune::Hybrid::ifElse(bool_t<visitChild>{}, [&] (auto /*id*/) -> void {
            applyToTree(child, childTP, visitor);
          });
          #endif // AMDIS_HAS_CXX_CONSTEXPR_IF
          visitor.afterChild(tree, child, treePath, i);
        });
        visitor.post(tree, treePath);
      }

      // Overload for leaf nodes
      template<class Tree, class TP, class Pre, class Leaf, class Post,
        std::enable_if_t<remove_cvref_t<Tree>::isLeaf, int> = 0>
      void for_each_node(Tree&& tree, TP treePath, Pre&& /*preFunc*/, Leaf&& leafFunc, Post&& /*postFunc*/)
      {
        leafFunc(tree, treePath);
      }

      // Overload for non-leaf nodes
      // Forward declaration needed for recursion
      template<class Tree, class TP, class Pre, class Leaf, class Post,
        std::enable_if_t<not remove_cvref_t<Tree>::isLeaf,int> = 0>
      void for_each_node(Tree&& tree, TP treePath, Pre&& preFunc, Leaf&& leafFunc, Post&& postFunc);

      // Helper for power nodes
      template<class Tree, class TP, class Pre, class Leaf, class Post, std::size_t... I,
        std::enable_if_t<remove_cvref_t<Tree>::isPower, int> = 0>
      void for_each_node_unfold(Tree&& tree, TP treePath, Pre&& preFunc, Leaf&& leafFunc, Post&& postFunc, std::index_sequence<I...>)
      {
        for (std::size_t i = 0; i < sizeof...(I); ++i)
          Impl::for_each_node(tree.child(i), Dune::TypeTree::push_back(treePath, i), preFunc, leafFunc, postFunc);
      }

      // Helper for composite nodes
      template<class Tree, class TP, class Pre, class Leaf, class Post, std::size_t... I,
        std::enable_if_t<not remove_cvref_t<Tree>::isPower, int> = 0>
      void for_each_node_unfold(Tree&& tree, TP treePath, Pre&& preFunc, Leaf&& leafFunc, Post&& postFunc, std::index_sequence<I...>)
      {
        (void)std::initializer_list<int>{(
          Impl::for_each_node(tree.child(Dune::index_constant<I>{}),
            Dune::TypeTree::push_back(treePath, Dune::index_constant<I>{}), preFunc, leafFunc, postFunc),0)...
        };
      }

      /*
       * Traverse tree and visit each node. The signature is the same
       * as for the public for_each_node function in Dune::Typtree,
       * despite the additionally passed treePath argument. The path
       * passed here is associated to the tree and the relative
       * paths of the children (wrt. to tree) are appended to this.
       * Hence the behavior of the public function is resembled
       * by passing an empty treePath.
       *
       * See also the specialization for leaf-nodes.
       */
      template<class Tree, class TP, class Pre, class Leaf, class Post,
        std::enable_if_t<not remove_cvref_t<Tree>::isLeaf,int>>
      void for_each_node(Tree&& tree, TP treePath, Pre&& preFunc, Leaf&& leafFunc, Post&& postFunc)
      {
        auto indices = std::make_index_sequence<TYPEOF(tree)::degree()>{};
        preFunc(tree, treePath);
        Impl::for_each_node_unfold(tree, treePath, preFunc, leafFunc, postFunc, indices);
        postFunc(tree, treePath);
      }

    } // namespace Impl


    // ********************************************************************************
    // Public Interface
    // ********************************************************************************

    //! Apply visitor to TypeTree.
    /**
     * \code
     * #include <amdis/typetree/Traversal.hpp>
     * \endcode
     * This function applies the given visitor to the given tree. Both visitor and tree may be const
     * or non-const (if the compiler supports rvalue references, they may even be a non-const temporary).
     *
     * \note The visitor must implement the interface laid out by DefaultVisitor (most easily achieved by
     *       inheriting from it) and specify the required type of tree traversal (static or dynamic) by
     *       inheriting from either StaticTraversal or DynamicTraversal.
     *
     * \param tree    The tree the visitor will be applied to.
     * \param visitor The visitor to apply to the tree.
     */
    template<typename Tree, typename Visitor>
    void apply_to_tree(Tree&& tree, Visitor&& visitor)
    {
      auto root = Dune::TypeTree::hybridTreePath();
      Impl::apply_to_tree(tree, root, visitor);
    }

    /**
     * \brief Traverse tree and visit each node
     *
     * All passed callback functions are called with the
     * node and corresponding treepath as arguments.
     *
     * \param tree The tree to traverse
     * \param preFunc This function is called for each inner node before visiting its children
     * \param leafFunc This function is called for each leaf node
     * \param postFunc This function is called for each inner node after visiting its children
     */
    template<class Tree, class Pre, class Leaf, class Post>
    void for_each_node(Tree&& tree, Pre&& preFunc, Leaf&& leafFunc, Post&& postFunc)
    {
      auto root = Dune::TypeTree::hybridTreePath();
      Impl::for_each_node(tree, root, preFunc, leafFunc, postFunc);
    }

    /**
     * \brief Traverse tree and visit each node
     *
     * All passed callback functions are called with the
     * node and corresponding treepath as arguments.
     *
     * \param tree The tree to traverse
     * \param innerFunc This function is called for each inner node before visiting its children
     * \param leafFunc This function is called for each leaf node
     */
    template<class Tree, class Inner, class Leaf>
    void for_each_node(Tree&& tree, Inner&& innerFunc, Leaf&& leafFunc)
    {
      auto root = Dune::TypeTree::hybridTreePath();
      Impl::for_each_node(tree, root, innerFunc, leafFunc, NoOp{});
    }

    /**
     * \brief Traverse tree and visit each node
     *
     * The passed callback function is called with the
     * node and corresponding treepath as arguments.
     *
     * \param tree The tree to traverse
     * \param nodeFunc This function is called for each node
     */
    template<class Tree, class NodeFunc>
    void for_each_node(Tree&& tree, NodeFunc&& nodeFunc)
    {
      auto root = Dune::TypeTree::hybridTreePath();
      Impl::for_each_node(tree, root, nodeFunc, nodeFunc, NoOp{});
    }

    /**
     * \brief Traverse tree and visit each leaf node
     *
     * The passed callback function is called with the
     * node and corresponding treepath as arguments.
     *
     * \param tree The tree to traverse
     * \param leafFunc This function is called for each leaf node
     */
    template<class Tree, class Leaf>
    void for_each_leaf_node(Tree&& tree, Leaf&& leafFunc)
    {
      auto root = Dune::TypeTree::hybridTreePath();
      Impl::for_each_node(tree, root, NoOp{}, leafFunc, NoOp{});
    }

} // end namespace AMDiS
