#include <iostream>
#include <memory>
#include <string>

#include <amdis/Observer.hpp>

#include "Tests.hpp"

namespace AMDiS {

namespace event {
  struct fooEvent {};
  struct barEvent {};
  struct bazEvent {};
}

} // end namespace AMDiS

std::string testValue = "";

bool checkAndResetTestValue(std::string testName, std::string ref, std::string refOpt = "X")
{
  bool result = testValue == ref || testValue == refOpt;
  testValue.clear();
  return result;
}

using namespace AMDiS;

class A
    : public Notifier<event::fooEvent, event::barEvent, event::bazEvent>
{
public:
  void foo()
  {
    std::cout << "A::foo()\n";
    testValue += "A.foo ";
    event::fooEvent e;
    this->notify(e);
  }

  void bar()
  {
    std::cout << "A::bar()\n";
    testValue += "A.bar ";
    event::barEvent e;
    this->notify(e);
  }

  void baz()
  {
    std::cout << "A::baz()\n";
    testValue += "A.baz ";
    event::bazEvent e;
    this->notify(e);
  }
};

class B
    : private Observer<event::fooEvent>
    , private Observer<event::bazEvent>
    , public Notifier<event::bazEvent>
{
public:
  B(std::shared_ptr<A> a)
    : Observer<event::fooEvent>(*a)
    , Observer<event::bazEvent>(*a)
    , a_(a)
  {}

  void updateImpl(event::fooEvent) override
  {
    std::cout << "B::update(foo)\n";
    testValue += "B.foo ";
  }

  void updateImpl(event::bazEvent e) override
  {
    std::cout << "B::update(baz)\n";
    testValue += "B.baz ";
    this->notify(e);
  }

  std::shared_ptr<A> const& a() const { return a_; }

  std::shared_ptr<A> a_;
};

class C
    : private Observer<event::barEvent>
    , private Observer<event::bazEvent>
{
public:
  C(std::shared_ptr<B> b)
    : Observer<event::barEvent>(*b->a())
    , Observer<event::bazEvent>(*b)
  {}

  void updateImpl(event::barEvent) override
  {
    std::cout << "C::update(bar)\n";
    testValue += "C.bar ";
  }

  void updateImpl(event::bazEvent) override
  {
    std::cout << "C::update(baz)\n";
    testValue += "C.baz ";
  }
};

class D
    : public Notifier<event::barEvent>
{
public:
  void bar()
  {
    std::cout << "D::bar()\n";
    testValue += "D.bar ";
    event::barEvent e;
    this->notify(e);
  }
};

class E
    : public Observer<event::fooEvent>
    , public Observer<event::barEvent>
{
public:
  E(std::shared_ptr<A> a, std::shared_ptr<D> d)
      : Observer<event::fooEvent>(*a)
      , Observer<event::barEvent>(*d)
  {}

  void updateImpl(event::fooEvent) override
  {
    std::cout << "E::update(foo)\n";
    testValue += "E.foo ";
  }

  void updateImpl(event::barEvent) override
  {
    std::cout << "E::update(bar)\n";
    testValue += "E.bar ";
  }
};

/// Observing a class that inherits from Observer<T, Events...> more than once is not supported
// class F
//     : public Observer<E, event::fooEvent>
// {
// public:
//   F(std::shared_ptr<E> e)
//     : Observer<E, event::fooEvent>(e)  // Error
//   {}
// };

/// Test the observer hierarchy.
/**
 *  There are 4 hierarchies tested.
 *  A::foo(): A -> B, E     (multiple observers [B,E] of one event)
 *  A::bar(): A -> (B) -> C (indirect observer [C of A])
 *  A::baz(): A -> B -> C   (simultanious observer and signaller of an event [B])
 *  D::bar(): D -> E        (observer of multiple classes [E])
 */
int main()
{
  using namespace AMDiS;

  auto a = std::make_shared<A>();
  auto b = std::make_shared<B>(a);
  auto c = std::make_shared<C>(b);
  auto d = std::make_shared<D>();
  auto e = std::make_shared<E>(a, d);
  std::string ref, refOpt;

  a->foo();
  ref = "A.foo B.foo E.foo ";
  refOpt = "A.foo E.foo B.foo ";  // Order of b->foo() and e->foo() unspecified
  AMDIS_TEST(checkAndResetTestValue("A:foo()", ref, refOpt));

  a->bar();
  ref = "A.bar C.bar ";
  AMDIS_TEST(checkAndResetTestValue("A:bar()", ref));

  a->baz();
  ref = "A.baz B.baz C.baz ";
  AMDIS_TEST(checkAndResetTestValue("A:baz()", ref));

  d->bar();
  ref = "D.bar E.bar ";
  AMDIS_TEST(checkAndResetTestValue("D:bar()", ref));

  return report_errors();
}
